Class: MutationTester::MutationRunner
- Inherits:
-
Object
- Object
- MutationTester::MutationRunner
- Defined in:
- lib/mutation_tester/mutation_runner.rb
Constant Summary collapse
- CANARY_SOURCE =
"raise 'mutation_tester canary: the workspace copy of this source was not executed'\n".freeze
Class Method Summary collapse
Instance Method Summary collapse
- #cleanup_shadow_workspaces ⇒ Object
- #detect_test_framework(spec_path) ⇒ Object
- #discoverable_project_root ⇒ Object
- #find_project_root ⇒ Object
- #in_memory_first? ⇒ Boolean
-
#initialize(source_file, spec_file, original_content, config) ⇒ MutationRunner
constructor
A new instance of MutationRunner.
- #run(mutations, &progress_callback) ⇒ Object
- #run_file_based_series(mutations, &progress_callback) ⇒ Object
- #run_in_memory_parallel(mutations, &progress_callback) ⇒ Object
- #run_in_memory_series(mutations, &progress_callback) ⇒ Object
- #run_in_place_series(mutations, &progress_callback) ⇒ Object
- #run_in_shadow_parallel(mutations, &progress_callback) ⇒ Object
- #run_in_shadow_series(mutations, project_root, &progress_callback) ⇒ Object
- #run_mutation_in_memory(mutation, result, project_root = nil) ⇒ Object
- #run_mutation_in_place(mutation, result) ⇒ Object
- #run_mutation_in_shadow(mutation, result, project_root) ⇒ Object
- #run_single_mutation(mutation, strategy, project_root = nil) ⇒ Object
- #run_specs_in_place(spec_file, example_filters: []) ⇒ Object
- #run_specs_in_shadow(spec_file, working_dir, project_root, example_filters: []) ⇒ Object
- #run_two_phase(mutation) ⇒ Object
- #shadow_baseline_passes? ⇒ Boolean
- #shadow_copy_project(source, dest) ⇒ Object
- #shadow_workspace_check ⇒ Object
- #subset_filters(mutation) ⇒ Object
Constructor Details
#initialize(source_file, spec_file, original_content, config) ⇒ MutationRunner
Returns a new instance of MutationRunner.
31 32 33 34 35 36 37 |
# File 'lib/mutation_tester/mutation_runner.rb', line 31 def initialize(source_file, spec_file, original_content, config) @source_file = File.(source_file) @spec_file = File.(spec_file) @original_content = original_content @config = config @use_bundle_exec = TestCommand.use_bundle_exec?(@source_file) end |
Class Method Details
.backup_path_for(source_file) ⇒ Object
39 40 41 |
# File 'lib/mutation_tester/mutation_runner.rb', line 39 def self.backup_path_for(source_file) "#{File.(source_file)}.mutation_backup" end |
.recover_in_place_backup(source_file) ⇒ Object
43 44 45 46 47 48 49 50 |
# File 'lib/mutation_tester/mutation_runner.rb', line 43 def self.recover_in_place_backup(source_file) backup = backup_path_for(source_file) return false unless File.exist?(backup) File.write(File.(source_file), File.read(backup)) File.delete(backup) true end |
Instance Method Details
#cleanup_shadow_workspaces ⇒ Object
258 259 260 261 262 263 264 265 266 267 268 269 |
# File 'lib/mutation_tester/mutation_runner.rb', line 258 def cleanup_shadow_workspaces root = @shadow_run_root return unless root begin FileUtils.remove_entry(root) if File.directory?(root) rescue SystemCallError nil end @shadow_run_root = nil @worker_shadow_root = nil end |
#detect_test_framework(spec_path) ⇒ Object
416 417 418 |
# File 'lib/mutation_tester/mutation_runner.rb', line 416 def detect_test_framework(spec_path) FrameworkDetector.detect(spec_path) end |
#discoverable_project_root ⇒ Object
329 330 331 332 333 |
# File 'lib/mutation_tester/mutation_runner.rb', line 329 def discoverable_project_root find_project_root rescue MutationTester::Error nil end |
#find_project_root ⇒ Object
335 336 337 338 339 340 341 342 343 344 345 346 347 348 |
# File 'lib/mutation_tester/mutation_runner.rb', line 335 def find_project_root current = File.dirname(File.(@source_file)) loop do return current if File.exist?(File.join(current, 'Gemfile')) || File.exist?(File.join(current, '.git')) parent = File.dirname(current) if parent == current || parent == Dir.home || parent == '/' raise MutationTester::Error, "Could not find project root (looking for Gemfile or .git). Stopped at #{parent}" end current = parent end end |
#in_memory_first? ⇒ Boolean
65 66 67 68 69 70 |
# File 'lib/mutation_tester/mutation_runner.rb', line 65 def in_memory_first? return false unless %i[in_memory auto].include?(@config.runner) return true unless @config.worker_env_var @config.parallel_processes == 1 || !@config.after_fork_file.nil? end |
#run(mutations, &progress_callback) ⇒ Object
52 53 54 55 56 57 58 59 60 61 62 63 |
# File 'lib/mutation_tester/mutation_runner.rb', line 52 def run(mutations, &progress_callback) announce_worker_env_in_memory_opt_out if in_memory_first? && @config.parallel_processes > 1 run_in_memory_parallel(mutations, &progress_callback) elsif in_memory_first? run_in_memory_series(mutations, &progress_callback) elsif @config.parallel_processes == 1 run_file_based_series(mutations, &progress_callback) else run_in_shadow_parallel(mutations, &progress_callback) end end |
#run_file_based_series(mutations, &progress_callback) ⇒ Object
170 171 172 173 174 175 |
# File 'lib/mutation_tester/mutation_runner.rb', line 170 def run_file_based_series(mutations, &progress_callback) project_root = @config.worker_env_var ? discoverable_project_root : nil return run_in_place_series(mutations, &progress_callback) unless project_root run_in_shadow_series(mutations, project_root, &progress_callback) end |
#run_in_memory_parallel(mutations, &progress_callback) ⇒ Object
100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 |
# File 'lib/mutation_tester/mutation_runner.rb', line 100 def run_in_memory_parallel(mutations, &progress_callback) blocker = prepare_in_memory_execution return fall_back_to_parallel_file_based(blocker, mutations, &progress_callback) if blocker pool = ForkRunner.prepare_in_memory_pool( [@config.parallel_processes, mutations.size].min, @in_memory_runner, env_for: worker_env_for, after_fork: @config.after_fork_file ) if pool.compact.empty? return fall_back_to_parallel_file_based('the preloaded worker pool could not be cloned', mutations, &progress_callback) end warn "[MutationTester] In-memory execution selected (parallel, #{pool.size} preloaded workers, zero file writes per mutant)." announce_load_time_routing(mutations) total = mutations.size completed_count = 0 collected = [] project_root = discoverable_project_root reserve_fallback_shadow_root report_progress = lambda do |_item, _index, result| completed_count += 1 collected << result if progress_callback && (completed_count.even? || completed_count == total) progress_callback.call(nil, completed_count) end raise Parallel::Break if stop_early?(result) end mapped = with_parallel_interrupt_silenced do Parallel.map(mutations, in_processes: pool.size, finish: report_progress) do |mutation| run_single_mutation(mutation, :in_memory, project_root) end end mapped || collected ensure ForkRunner.shutdown_in_memory_pool shutdown_in_memory_worker cleanup_shadow_workspaces end |
#run_in_memory_series(mutations, &progress_callback) ⇒ Object
72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 94 95 96 97 98 |
# File 'lib/mutation_tester/mutation_runner.rb', line 72 def run_in_memory_series(mutations, &progress_callback) blocker = prepare_in_memory_execution return fall_back_to_file_based(blocker, mutations, &progress_callback) if blocker warn '[MutationTester] In-memory execution selected (serial, zero file writes per mutant).' announce_load_time_routing(mutations) results = [] begin mutations.each_with_index do |mutation, index| unless @in_memory_runner&.ready? return results + fall_back_to_file_based( 'the in-memory worker terminated unexpectedly', mutations.drop(index), completed: index, &progress_callback ) end result = run_single_mutation(mutation, :in_memory) progress_callback.call(mutation, index + 1) if progress_callback results << result break if stop_early?(result) end ensure shutdown_in_memory_worker cleanup_shadow_workspaces end results end |
#run_in_place_series(mutations, &progress_callback) ⇒ Object
192 193 194 195 196 197 198 199 200 201 202 203 204 |
# File 'lib/mutation_tester/mutation_runner.rb', line 192 def run_in_place_series(mutations, &progress_callback) write_in_place_backup results = [] mutations.each_with_index do |mutation, index| result = run_single_mutation(mutation, :in_place) progress_callback.call(mutation, index + 1) if progress_callback results << result break if stop_early?(result) end results ensure restore_and_clear_in_place_backup end |
#run_in_shadow_parallel(mutations, &progress_callback) ⇒ Object
143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 |
# File 'lib/mutation_tester/mutation_runner.rb', line 143 def run_in_shadow_parallel(mutations, &progress_callback) total = mutations.size completed_count = 0 collected = [] project_root = find_project_root shadow_run_root prepare_worker_preloads(total) report_progress = lambda do |_item, _index, result| completed_count += 1 collected << result if progress_callback && (completed_count.even? || completed_count == total) progress_callback.call(nil, completed_count) end raise Parallel::Break if stop_early?(result) end mapped = with_parallel_interrupt_silenced do Parallel.map(mutations, in_processes: @config.parallel_processes, finish: report_progress) do |mutation| run_single_mutation(mutation, :shadow, project_root) end end mapped || collected ensure cleanup_shadow_workspaces end |
#run_in_shadow_series(mutations, project_root, &progress_callback) ⇒ Object
177 178 179 180 181 182 183 184 185 186 187 188 189 190 |
# File 'lib/mutation_tester/mutation_runner.rb', line 177 def run_in_shadow_series(mutations, project_root, &progress_callback) warn "[MutationTester] --worker-env #{@config.worker_env_var} is set, so the serial run decides each mutant in a shadow workspace and never mutates the checkout in place." shadow_run_root results = [] mutations.each_with_index do |mutation, index| result = run_single_mutation(mutation, :shadow, project_root) progress_callback.call(mutation, index + 1) if progress_callback results << result break if stop_early?(result) end results ensure cleanup_shadow_workspaces end |
#run_mutation_in_memory(mutation, result, project_root = nil) ⇒ Object
369 370 371 372 373 374 375 376 377 378 379 380 381 382 383 384 385 386 387 388 389 390 391 |
# File 'lib/mutation_tester/mutation_runner.rb', line 369 def run_mutation_in_memory(mutation, result, project_root = nil) runner = active_in_memory_runner return in_memory_worker_fallback(mutation, result, project_root, 'the in-memory worker is unavailable') unless runner&.ready? outcome = runner.execute_in_memory( source: mutation[:code], path: @source_file, timeout: @config.effective_timeout, chdir: Dir.pwd ) if outcome.status == 'error' in_memory_apply_fallback(mutation, result, project_root, outcome. || 'in-memory application failed') else apply_outcome(result, TestCommand::Result.new(outcome.status == 'pass', outcome.status == 'timeout')) end rescue MutationTester::Error => e return in_memory_worker_fallback(mutation, result, project_root, e.) if project_root mark_error(result, e) rescue StandardError => e mark_error(result, e) end |
#run_mutation_in_place(mutation, result) ⇒ Object
350 351 352 353 354 355 356 357 358 359 360 361 362 363 |
# File 'lib/mutation_tester/mutation_runner.rb', line 350 def run_mutation_in_place(mutation, result) File.write(@source_file, mutation[:code]) outcome, phase = run_two_phase(mutation) do |example_filters| run_specs_in_place(@spec_file, example_filters: example_filters) end apply_outcome(result, outcome, phase) rescue StandardError, Interrupt => e mark_error(result, e) raise e if e.is_a?(Interrupt) ensure File.write(@source_file, @original_content) end |
#run_mutation_in_shadow(mutation, result, project_root) ⇒ Object
235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256 |
# File 'lib/mutation_tester/mutation_runner.rb', line 235 def run_mutation_in_shadow(mutation, result, project_root) shadow_root = worker_shadow_root(project_root) relative_source = Pathname.new(@source_file).relative_path_from(Pathname.new(project_root)).to_s shadow_source = File.join(shadow_root, relative_source) relative_spec = Pathname.new(@spec_file).relative_path_from(Pathname.new(project_root)).to_s shadow_spec = File.join(shadow_root, relative_spec) begin File.unlink(shadow_source) if File.symlink?(shadow_source) File.write(shadow_source, mutation[:code]) outcome, phase = run_two_phase(mutation) do |example_filters| run_specs_in_shadow(shadow_spec, shadow_root, project_root, example_filters: example_filters) end apply_outcome(result, outcome, phase) ensure restore_shadow_source(shadow_source, result) end rescue => e mark_error(result, e) end |
#run_single_mutation(mutation, strategy, project_root = nil) ⇒ Object
206 207 208 209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 233 |
# File 'lib/mutation_tester/mutation_runner.rb', line 206 def run_single_mutation(mutation, strategy, project_root = nil) { id: mutation[:id], type: mutation[:type], line: mutation[:line], file_path: @source_file, original: mutation[:original], mutated: mutation[:mutated], source_line: mutation[:source_line], mutated_line: mutation[:mutated_line], killed: false, timeout: false, status: :survived, description: mutation[:description] }.tap do |result| if unparseable?(mutation[:code]) mark_stillborn(result) elsif strategy == :in_memory && mutation[:in_memory_safe] == false run_mutation_load_time(mutation, result, project_root) elsif strategy == :in_memory run_mutation_in_memory(mutation, result, project_root) elsif strategy == :in_place run_mutation_in_place(mutation, result) else run_mutation_in_shadow(mutation, result, project_root) end end end |
#run_specs_in_place(spec_file, example_filters: []) ⇒ Object
365 366 367 |
# File 'lib/mutation_tester/mutation_runner.rb', line 365 def run_specs_in_place(spec_file, example_filters: []) test_command(spec_file, example_filters: example_filters).run(timeout: @config.effective_timeout) end |
#run_specs_in_shadow(spec_file, working_dir, project_root, example_filters: []) ⇒ Object
324 325 326 327 |
# File 'lib/mutation_tester/mutation_runner.rb', line 324 def run_specs_in_shadow(spec_file, working_dir, project_root, example_filters: []) test_command(spec_file, example_filters: example_filters) .run(timeout: @config.effective_timeout, chdir: working_dir, mirror_of: project_root) end |
#run_two_phase(mutation) ⇒ Object
393 394 395 396 397 398 399 400 401 |
# File 'lib/mutation_tester/mutation_runner.rb', line 393 def run_two_phase(mutation) filters = subset_filters(mutation) unless filters.empty? subset_outcome = yield(filters) return [subset_outcome, :subset] unless subset_outcome.passed? end [yield([]), :full] end |
#shadow_baseline_passes? ⇒ Boolean
271 272 273 |
# File 'lib/mutation_tester/mutation_runner.rb', line 271 def shadow_baseline_passes? shadow_workspace_check == :ok end |
#shadow_copy_project(source, dest) ⇒ Object
302 303 304 305 306 307 308 309 310 311 312 313 314 315 316 317 318 319 320 321 322 |
# File 'lib/mutation_tester/mutation_runner.rb', line 302 def shadow_copy_project(source, dest) if source == '/' || source.match?(%r{^/(usr|bin|sbin|etc|var|opt)$}) raise MutationTester::Error, "Refusing to shadow copy system directory: #{source}" end Dir.glob("#{source}/*", File::FNM_DOTMATCH).each do |path| next if ['.', '..', '.git', 'tmp', 'log', 'coverage', 'node_modules'].include?(File.basename(path)) basename = File.basename(path) target = File.join(dest, basename) if File.directory?(path) && !File.symlink?(path) FileUtils.mkdir_p(target) shadow_copy_project(path, target) elsif File.extname(path) == '.rb' FileUtils.copy_file(path, target) else File.symlink(path, target) end end end |
#shadow_workspace_check ⇒ Object
275 276 277 278 279 280 281 282 283 284 285 286 287 288 289 290 291 292 293 294 295 296 297 298 299 300 |
# File 'lib/mutation_tester/mutation_runner.rb', line 275 def shadow_workspace_check project_root = find_project_root Dir.mktmpdir do |temp_dir| shadow_root = File.join(temp_dir, 'shadow') FileUtils.mkdir_p(shadow_root) shadow_copy_project(project_root, shadow_root) relative_source = Pathname.new(@source_file).relative_path_from(Pathname.new(project_root)).to_s shadow_source = File.join(shadow_root, relative_source) relative_spec = Pathname.new(@spec_file).relative_path_from(Pathname.new(project_root)).to_s shadow_spec = File.join(shadow_root, relative_spec) File.unlink(shadow_source) File.write(shadow_source, @original_content) return :baseline unless run_specs_in_shadow(shadow_spec, shadow_root, project_root).passed? File.write(shadow_source, CANARY_SOURCE) return :canary if run_specs_in_shadow(shadow_spec, shadow_root, project_root).passed? :ok end rescue => e warn("[MutationTester] Shadow sanity check could not prepare the shadow workspace: #{e.}") :baseline end |
#subset_filters(mutation) ⇒ Object
403 404 405 406 407 408 409 410 411 412 413 414 |
# File 'lib/mutation_tester/mutation_runner.rb', line 403 def subset_filters(mutation) return [] unless @config.test_selection return [] unless detect_test_framework(@spec_file) == :rspec method_name = mutation[:method_name].to_s return [] if method_name.empty? content = spec_content ["##{method_name}", ".#{method_name}"].select do |token| content.match?(/['"]#{Regexp.escape(token)}/) end end |